import os
import json
from tempfile import mkstemp

import numpy as np
import scipy.sparse as ss

import h5sparse


def test_create_and_read_dataset():
    h5_path = mkstemp(suffix=".h5")[1]
    sparse_matrix = ss.csr_matrix([[0, 1, 0],
                                   [0, 0, 1],
                                   [0, 0, 0],
                                   [1, 1, 0]],
                                  dtype=np.float64)
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('sparse/matrix', data=sparse_matrix)
    with h5sparse.File(h5_path) as h5f:
        assert 'sparse' in h5f
        assert 'matrix' in h5f['sparse']
        assert (h5f['sparse']['matrix'][1:3] != sparse_matrix[1:3]).size == 0
        assert (h5f['sparse']['matrix'][2:] != sparse_matrix[2:]).size == 0
        assert (h5f['sparse']['matrix'][:2] != sparse_matrix[:2]).size == 0
        assert (h5f['sparse']['matrix'][-2:] != sparse_matrix[-2:]).size == 0
        assert (h5f['sparse']['matrix'][:-2] != sparse_matrix[:-2]).size == 0
        assert (h5f['sparse']['matrix'][()] != sparse_matrix).size == 0

    os.remove(h5_path)


def test_create_dataset_with_format_change():
    h5_path = mkstemp(suffix=".h5")[1]
    sparse_matrix = ss.csr_matrix([[0, 1, 0, 1],
                                   [0, 0, 1, 0],
                                   [0, 0, 0, 1],
                                   [1, 1, 0, 1]],
                                  dtype=np.float64)
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('sparse/matrix', data=sparse_matrix, sparse_format='csc')
    with h5sparse.File(h5_path) as h5f:
        assert 'sparse' in h5f
        assert 'matrix' in h5f['sparse']
        assert h5f['sparse']['matrix'].format_str == 'csc'
        result_matrix = h5f['sparse']['matrix'][()]
        assert isinstance(result_matrix, ss.csc_matrix)
        assert (result_matrix != sparse_matrix).size == 0
        assert (h5f['sparse']['matrix'][1:3] != sparse_matrix[:, 1:3]).size == 0
        assert (h5f['sparse']['matrix'][2:] != sparse_matrix[:, 2:]).size == 0
        assert (h5f['sparse']['matrix'][:2] != sparse_matrix[:, :2]).size == 0
        assert (h5f['sparse']['matrix'][-2:] != sparse_matrix[:, -2:]).size == 0
        assert (h5f['sparse']['matrix'][:-2] != sparse_matrix[:, :-2]).size == 0

    os.remove(h5_path)


def test_create_empty_sparse_dataset():
    h5_path = mkstemp(suffix=".h5")[1]
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('sparse/matrix', sparse_format='csr')
    with h5sparse.File(h5_path) as h5f:
        assert 'sparse' in h5f
        assert 'matrix' in h5f['sparse']
        assert h5f['sparse']['matrix'].format_str == 'csr'
        result_matrix = h5f['sparse']['matrix'][()]
        assert isinstance(result_matrix, ss.csr_matrix)
        assert result_matrix.shape == (0, 0)
        assert result_matrix.dtype == np.float64
        assert h5f['sparse']['matrix'].shape == (0, 0)
        assert h5f['sparse']['matrix'].dtype == np.float64

    os.remove(h5_path)


def test_create_dataset_from_dataset():
    from_h5_path = mkstemp(suffix=".h5")[1]
    to_h5_path = mkstemp(suffix=".h5")[1]
    sparse_matrix = ss.csr_matrix([[0, 1, 0],
                                   [0, 0, 1],
                                   [0, 0, 0],
                                   [1, 1, 0]],
                                  dtype=np.float64)
    with h5sparse.File(from_h5_path) as from_h5f:
        from_dset = from_h5f.create_dataset('sparse/matrix', data=sparse_matrix)

        with h5sparse.File(to_h5_path) as to_h5f:
            to_h5f.create_dataset('sparse/matrix', data=from_dset)
            assert 'sparse' in to_h5f
            assert 'matrix' in to_h5f['sparse']
            assert (to_h5f['sparse/matrix'][()] != sparse_matrix).size == 0

    os.remove(from_h5_path)
    os.remove(to_h5_path)


def test_dataset_append():
    h5_path = mkstemp(suffix=".h5")[1]
    sparse_matrix = ss.csr_matrix([[0, 1, 0],
                                   [0, 0, 1],
                                   [0, 0, 0],
                                   [1, 1, 0]],
                                  dtype=np.float64)
    to_append = ss.csr_matrix([[0, 1, 1],
                               [1, 0, 0]],
                              dtype=np.float64)
    appended_matrix = ss.vstack((sparse_matrix, to_append))

    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('matrix', data=sparse_matrix, chunks=(100000,),
                           maxshape=(None,))
        h5f['matrix'].append(to_append)
        assert (h5f['matrix'][()] != appended_matrix).size == 0

    os.remove(h5_path)


def test_numpy_array():
    h5_path = mkstemp(suffix=".h5")[1]
    matrix = np.random.rand(3, 5)
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('matrix', data=matrix)
        assert 'matrix' in h5f
        np.testing.assert_equal(h5f['matrix'][()], matrix)
    os.remove(h5_path)


def test_bytestring():
    h5_path = mkstemp(suffix=".h5")[1]
    strings = [str(i) for i in range(100)]
    data = json.dumps(strings).encode('utf8')
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('strings', data=data)
        assert 'strings' in h5f
        assert strings == json.loads(h5f['strings'][()].decode('utf8'))
    os.remove(h5_path)


def test_create_empty_dataset():
    h5_path = mkstemp(suffix=".h5")[1]
    with h5sparse.File(h5_path) as h5f:
        h5f.create_dataset('empty_data', shape=(100, 200))
    with h5sparse.File(h5_path) as h5f:
        assert h5f['empty_data'].shape == (100, 200)
    os.remove(h5_path)



import six
import h5py
import numpy as np
import scipy.sparse as ss


FORMAT_DICT = {
    'csr': ss.csr_matrix,
    'csc': ss.csc_matrix,
}


def get_format_str(data):
    for format_str, format_class in six.viewitems(FORMAT_DICT):
        if isinstance(data, format_class):
            return format_str
    raise ValueError("Data type {} is not supported.".format(type(data)))


def get_format_class(format_str):
    try:
        format_class = FORMAT_DICT[format_str]
    except KeyError:
        raise ValueError("Format {} is not supported."
                         .format(format_str))
    return format_class


class Group(h5py.Group):
    """The HDF5 group that can detect and create sparse matrix.
    """

    def __getitem__(self, key):
        h5py_item = super(Group, self).__getitem__(key)
        if isinstance(h5py_item, h5py.Group):
            if 'h5sparse_format' in h5py_item.attrs:
                # detect the sparse matrix
                return Dataset(h5py_item)
            else:
                return Group(h5py_item.id)
        elif isinstance(h5py_item, h5py.Dataset):
            return h5py_item
        else:
            raise ValueError("Unexpected item type.")

    def create_dataset(self, name, shape=None, dtype=None, data=None,
                       sparse_format=None, indptr_dtype=np.int64, indices_dtype=np.int32,
                       **kwargs):
        """Create 3 datasets in a group to represent the sparse array.

        Parameters
        ----------
        sparse_format:
        """
        if isinstance(data, Dataset):
            assert sparse_format is None
            group = self.create_group(name)
            group.attrs['h5sparse_format'] = data.attrs['h5sparse_format']
            group.attrs['h5sparse_shape'] = data.attrs['h5sparse_shape']
            group.create_dataset('data', data=data.h5py_group['data'],
                                 dtype=dtype, **kwargs)
            group.create_dataset('indices', data=data.h5py_group['indices'],
                                 dtype=indices_dtype, **kwargs)
            group.create_dataset('indptr', data=data.h5py_group['indptr'],
                                 dtype=indptr_dtype, **kwargs)
        elif ss.issparse(data):
            if sparse_format is not None:
                format_class = get_format_class(sparse_format)
                data = format_class(data)
            group = self.create_group(name)
            group.attrs['h5sparse_format'] = get_format_str(data)
            group.attrs['h5sparse_shape'] = data.shape
            group.create_dataset('data', data=data.data, dtype=dtype, **kwargs)
            group.create_dataset('indices', data=data.indices, dtype=indices_dtype, **kwargs)
            group.create_dataset('indptr', data=data.indptr, dtype=indptr_dtype, **kwargs)
        elif data is None and sparse_format is not None:
            format_class = get_format_class(sparse_format)
            if dtype is None:
                dtype = np.float64
            if shape is None:
                shape = (0, 0)
            data = format_class(shape, dtype=dtype)
            group = self.create_group(name)
            group.attrs['h5sparse_format'] = get_format_str(data)
            group.attrs['h5sparse_shape'] = data.shape
            group.create_dataset('data', data=data.data, dtype=dtype, **kwargs)
            group.create_dataset('indices', data=data.indices, dtype=indices_dtype, **kwargs)
            group.create_dataset('indptr', data=data.indptr, dtype=indptr_dtype, **kwargs)
        else:
            # forward the arguments to h5py
            assert sparse_format is None
            return super(Group, self).create_dataset(
                name, data=data, shape=shape, dtype=dtype, **kwargs)
        return Dataset(group)


class File(h5py.File, Group):
    """The HDF5 file object that can detect and create sparse matrix.
    """
    pass


class Dataset(h5py.Group):
    """The HDF5 sparse matrix dataset.

    Parameters
    ----------
    h5py_group : h5py.Dataset
    """

    def __init__(self, h5py_group):
        super(Dataset, self).__init__(h5py_group.id)
        self.h5py_group = h5py_group
        self.shape = tuple(self.attrs['h5sparse_shape'])
        self.format_str = self.attrs['h5sparse_format']
        self.dtype = h5py_group['data'].dtype
        self.indptr_dtype = h5py_group['indptr'].dtype
        self.indices_dtype = h5py_group['indices'].dtype

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise NotImplementedError("Index step is not supported.")
            start = key.start
            stop = key.stop
            if stop is not None and stop > 0:
                stop += 1
            if start is not None and start < 0:
                start -= 1
            indptr_slice = slice(start, stop)
            indptr = self.h5py_group['indptr'][indptr_slice]
            data = self.h5py_group['data'][indptr[0]:indptr[-1]]
            indices = self.h5py_group['indices'][indptr[0]:indptr[-1]]
            indptr -= indptr[0]
            if self.format_str == 'csr':
                shape = (indptr.size - 1, self.shape[1])
            elif self.format_str == 'csc':
                shape = (self.shape[0], indptr.size - 1)
            else:
                raise NotImplementedError("Slicing for format {} is not implemented."
                                          .format(self.format_str))
        elif isinstance(key, tuple) and key == ():
            data = self.h5py_group['data'][()]
            indices = self.h5py_group['indices'][()]
            indptr = self.h5py_group['indptr'][()]
            shape = self.shape
        else:
            raise NotImplementedError("Only support one slice as index.")

        format_class = get_format_class(self.attrs['h5sparse_format'])
        return format_class((data, indices, indptr), shape=shape)

    @property
    def value(self):
        return self[()]

    def append(self, sparse_matrix):
        if self.format_str != get_format_str(sparse_matrix):
            raise ValueError("Format not the same.")

        if self.format_str == 'csr':
            # data
            data = self.h5py_group['data']
            orig_data_size = data.shape[0]
            new_shape = (orig_data_size + sparse_matrix.data.shape[0],)
            data.resize(new_shape)
            data[orig_data_size:] = sparse_matrix.data

            # indptr
            indptr = self.h5py_group['indptr']
            orig_data_size = indptr.shape[0]
            append_offset = indptr[-1]
            new_shape = (orig_data_size + sparse_matrix.indptr.shape[0] - 1,)
            indptr.resize(new_shape)
            indptr[orig_data_size:] = (sparse_matrix.indptr[1:].astype(np.int64)
                                       + append_offset)

            # indices
            indices = self.h5py_group['indices']
            orig_data_size = indices.shape[0]
            new_shape = (orig_data_size + sparse_matrix.indices.shape[0],)
            indices.resize(new_shape)
            indices[orig_data_size:] = sparse_matrix.indices

            # shape
            self.shape = (
                self.shape[0] + sparse_matrix.shape[0],
                max(self.shape[1], sparse_matrix.shape[1]))
            self.attrs['h5sparse_shape'] = self.shape
        else:
            raise NotImplementedError("The append method for format {} is not implemented."
                                      .format(self.format_str))



#!/usr/bin/env python
from setuptools import setup


setup_requires = [
    'nose',
    'coverage',
]
install_requires = [
    'h5py',
    'numpy',
    'scipy',
    'six',
]
tests_require = []

description = "Scipy sparse matrix in HDF5."

long_description = """\
Please visit  the `Github repository <https://github.com/appier/h5sparse>`_
for more information.\n
"""
with open('README.rst') as fp:
    long_description += fp.read()


setup(
    name='h5sparse',
    version="0.1.0",
    description=description,
    long_description=long_description,
    author='Appier Inc.',
    url='https://github.com/appier/h5sparse',
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    license="MIT",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
    test_suite='nose.collector',
    packages=[
        'h5sparse',
    ],
)
